/*
CoryBiConnect4
by Cory McKay

CLASS NAME:		Analyzer
DESCRIPTION:	This class deals with the actual background work of the systen.
				The computer chooses a move using a MiniMax search strategy with alpha-beta pruning.
				The evaluation function is described in the report. Essentially, it gives points for
				the number of sets of consecutive pieces for the player, and subtracts points for the
				number of sets of the opponent. Sets of three are weighted four times as high as sets
				of two.
*/

class Analyzer
{
	/* PUBLIC METHODS **********************************************************/


	public char get_computer_move(char[][] board, char computer_colour, int turn)
	// Returns the computer's move based on the given board.
	// computer_colour is 'w' if the computer is playing white and 'b' otherwise
	{
		if (turn == 1)
			return '3'; // automatically take move 3 if moving first
		else
		{
			int max_depth = 6; // maximum depth of tree
			char [][] temp_board = copy_board(board); // current board
			int score = evaluate_board(temp_board, computer_colour); // goodness of current board
			ScoreNode results = max_value(temp_board, score, 'A', computer_colour, 0, max_depth, -20000, 20000);
			char move = check_legal_move(board, results.move);
			return move;
		}
	}


	public char is_game_over(char[][] board)
	// Returns 'w' if the games represented on board is over (i.e. 4 consecutive tokens
	// of the same colour are present) and has been won by white, returns 'b' if has been
	// won by black, returns 't' if it's a tie game and returns 'c' if game is not over yet.
	{
		// Check if it is a time game
		int count = 0; // number of top edge and right edge cells filled
		for (int i = 0; i < 7; i++)
		{
			if (board[i][0] != ',') count++;
			if (board[6][i] != ',') count++;
		}
		if (count == 14)
			return 't';

		// Check to see if there is a winner
		int[] results = count_contiguous(board, 'w'); // white arbitrarily chosen
		if (results[2] != 0)
			return 'w';
		else if (results[5] != 0)
			return 'b';
		else return 'c';
	}


	public void revise_state(char move, char[][] board, char colour)
	// Updates the given board after the given move by the given colour
	// A move in a column or row that is already full results in no change
	{
		int row = -1; // row that piece will be inserted into
		int column = -1; // column that piece will be inserted into
		char symbol; // symbol to be inserted on board

		// Translate colour
		if (colour == 'w') symbol = '0';
		else symbol = '1';

		// Translate move
		if (move == '0') column = 0;
		else if (move == '1') column = 1;
		else if (move == '2') column = 2;
		else if (move == '3') column = 3;
		else if (move == '4') column = 4;
		else if (move == '5') column = 5;
		else if (move == '6') column = 6;
		else if (move == 'a') row = 0;
		else if (move == 'b') row = 1;
		else if (move == 'c') row = 2;
		else if (move == 'd') row = 3;
		else if (move == 'e') row = 4;
		else if (move == 'f') row = 5;
		else row = 6;

		// Slide in piece inserted vertically
		if (row == -1)
		{
			row = 6;
			for (int i = 0; i < 7; i++)
			{
				if (board[column][i] != ',')
				{
					row = i - 1;
					i = 7;
				}
			}
		}

		// Slide in piece inserted horizontally
		else
		{
			column = 0;
			for (int i = 6; i >= 0; i--)
			{
				if (board[i][row] != ',')
				{
					column = i + 1;
					i = -1;
				}
			}
		}

		// Update the board if a piece was slid into an empty spot
		if (row >= 0 && column <= 6)
			board[column][row] = symbol;
	}


	/* PRIVATE METHODS *********************************************************/

	
	private ScoreNode max_value(char[][] board, int score, char move, char computer_colour,
	                            int depth, int max_depth, int alpha, int beta)
	// Max part of alpha-beta search algorithm
	{
		// Return evaluation if have reached cutoff depth
		if (depth == max_depth) return new ScoreNode(score, move);

		// Return the score if this state would result in a game over
		else if (score <= -10000 || score >= 10000) return new ScoreNode(score, move);

		// Return alpha
		else
		{
			// Spawn and sort children
			Object[] children_boards = new Object[14]; // potential state spaces
			int[] children_scores = new int[14];  // scores of each potential state space
			char[] moves = new char[14]; // moves leading to each potential state space
			set_up_children(children_boards, children_scores, moves, board, computer_colour, true);

			// Choose a default move if this is the first set of choices
			if (depth == 0)
			{
				int temp_index = 0;
				for (int i = 0; i < 14; i++)
				{
					if (children_scores[i] > children_scores[temp_index])
						temp_index = i;
				}
				move = moves[temp_index];
			}

			// Explore children and return appropriate value
			for (int i = 0; i < 14; i++)
			{
				ScoreNode min = min_value((char[][]) children_boards[i], children_scores[i], moves[i],
				                         computer_colour, depth + 1, max_depth, alpha, beta);
				if (min.score > alpha)
				{
					alpha = min.score;
					if (depth == 0)
						move = min.move; // best move to take so far
				}
				if (alpha >= beta) return new ScoreNode(beta, move);
			}
			return new ScoreNode(alpha, move);
		}
	}


	private ScoreNode min_value(char[][] board, int score, char move, char computer_colour,
	                            int depth, int max_depth, int alpha, int beta)
	// Min part of alpha-beta search algorithm
	{
		// Return evaluation if have reached cutoff depth
		if (depth == max_depth) return new ScoreNode(score, move);
		
		// Return the score if this state would result in a game over
		else if (score <= -10000 || score >= 10000) return new ScoreNode(score, move);

		// Return alpha
		else
		{
			// Spawn and sort children
			Object[] children_boards = new Object[14]; // potential state spaces
			int[] children_scores = new int[14];  // scores of each potential state space
			char[] moves = new char[14]; // moves leading to each potential state space
			set_up_children(children_boards, children_scores, moves, board, computer_colour, false);

			// Explore children and return appropriatevalue
			for (int i = 0; i < 14; i++)
			{
				int max = max_value((char[][]) children_boards[i], children_scores[i], moves[i], 
				                    computer_colour, depth + 1, max_depth, alpha, beta).score;
				if (max < beta) beta = max;
				if (beta <= alpha) return new ScoreNode(alpha, move);
			}
			return new ScoreNode(beta, move);
		}
	}


	private int evaluate_board(char[][] board, char playing_colour)
	// Returns a score indicating the goodness of the given board from the perspective of
	// the player playing the colour player_colour. The higher the score, the better move
	// for the player playing playing_colour.
	{
		int score = 0; // the goodness of the board from the perspective of playing_colour
		int[] counts = count_contiguous(board, playing_colour); // data on contiguous pieces

		// Always choose if you win and avoid if you lose
		score += 15000 * counts[2];
		score += -15000 * counts[5];

		// Strong bias towards you getting triples and against oppenent getting triples
		score += 20 * counts[1];
		score += -20 * counts[4];

		// Strong bias towards you getting doubles and against oppenent getting doubles
		score += 5 * counts[0];
		score += -5 * counts[3];

		return score;
	}


	private int[] count_contiguous(char[][] board, char playing_colour)
	// Counts the number of doubles, triples and quadruples for each colour on the given
	// board. Returns this information in a 6-cell array with the folowing format (where
	// pc means playing colour and "3's" means sets of 3 contiguous blocks of same colour):
	// <# of pc 2s><# of pc 3s><# of pc 4s><# of non-pc 2s><# of non-pc 3s><# of non-pc 4s>
	{
		int[] results = new int[6]; // counts of contiguous blocks
		int black_count; // number of black blocks found so far
		int white_count; // number of white blocks found so far
		int diag_length; // length of a diagonal
		int start_col; // first column of a diagonal

		// Initialize results
		for (int i = 0; i < 6; i++)
			results[i] = 0;

		// Count rows
		black_count = 0;
		white_count = 0;
		for (int i = 0; i < 7; i++)
		{
			for (int j = 0; j < 7; j++)
			{
				if (board[j][i] == ',')
				{
					if (white_count != 0)
					{
						update_count(results, white_count, black_count);
						white_count = 0;
					}
					else if (black_count != 0)
					{
						update_count(results, white_count, black_count);
						black_count = 0;
					}
				}
				else if (board[j][i] == '0')
				{
					if (black_count != 0)
					{
						update_count(results, white_count, black_count);
						black_count = 0;
					}
					white_count++;
				}
				else
				{
					if (white_count != 0)
					{
						update_count(results, white_count, black_count);
						white_count = 0;
					}
					black_count++;
				}

				if (j == 6 && board[j][i] != ',')
				{
					update_count(results, white_count, black_count);
					white_count = 0;
					black_count = 0;
				}
			}
		}

		// Count columns
		black_count = 0;
		white_count = 0;
		for (int j = 0; j < 7; j++)
		{
			for (int i = 0; i < 7; i++)
			{
				if (board[j][i] == ',')
				{
					if (white_count != 0)
					{
						update_count(results, white_count, black_count);
						white_count = 0;
					}
					else if (black_count != 0)
					{
						update_count(results, white_count, black_count);
						black_count = 0;
					}
				}
				else if (board[j][i] == '0')
				{
					if (black_count != 0)
					{
						update_count(results, white_count, black_count);
						black_count = 0;
					}
					white_count++;
				}
				else
				{
					if (white_count != 0)
					{
						update_count(results, white_count, black_count);
						white_count = 0;
					}
					black_count++;
				}

				if (i == 6 && board[j][i] != ',')
				{
					update_count(results, white_count, black_count);
					white_count = 0;
					black_count = 0;
				}
			}
		}
		
		// Count descending (as read from left to right) diagonals
		black_count = 0;
		white_count = 0;
		int i; // row index
		for (int diag = -5; diag < 6; diag++)
		{
			if (diag > 0)
			{
				diag_length = 7 - diag;
				start_col = diag;
				i = 0; // row index
			}
			else
			{
				diag_length = 7 + diag;
				start_col = 0;
				i = (-1) * diag; // row index
			}			

			for (int j = start_col; j < start_col + diag_length; j++) // j is the column index
			{
				if (board[j][i] == ',')
				{
					if (white_count != 0)
					{
						update_count(results, white_count, black_count);
						white_count = 0;
					}
					else if (black_count != 0)
					{
						update_count(results, white_count, black_count);
						black_count = 0;
					}
				}
				else if (board[j][i] == '0')
				{
					if (black_count != 0)
					{
						update_count(results, white_count, black_count);
						black_count = 0;
					}
					white_count++;
				}
				else
				{
					if (white_count != 0)
					{
						update_count(results, white_count, black_count);
						white_count = 0;
					}
					black_count++;
				}

				if ((j == 6 || i == 6) && board[j][i] != ',')
				{
					update_count(results, white_count, black_count);
					white_count = 0;
					black_count = 0;
				}
				
				i++;
			}
		}

		// Count ascending (as read from left to right) diagonals
		black_count = 0;
		white_count = 0;
		for (int diag = -5; diag < 6; diag++)
		{
			if (diag > 0)
			{
				diag_length = 7 - diag;
				start_col = diag;
				i = 6; // row index
			}
			else
			{
				diag_length = 7 + diag;
				start_col = 0;
				i = 6 + diag; // row index
			}			

			for (int j = start_col; j < start_col + diag_length; j++) // j is the column index
			{
				if (board[j][i] == ',')
				{
					if (white_count != 0)
					{
						update_count(results, white_count, black_count);
						white_count = 0;
					}
					else if (black_count != 0)
					{
						update_count(results, white_count, black_count);
						black_count = 0;
					}
				}
				else if (board[j][i] == '0')
				{
					if (black_count != 0)
					{
						update_count(results, white_count, black_count);
						black_count = 0;
					}
					white_count++;
				}
				else
				{
					if (white_count != 0)
					{
						update_count(results, white_count, black_count);
						white_count = 0;
					}
					black_count++;
				}

				if ((j == 6 || i == 0) && board[j][i] != ',')
				{
					update_count(results, white_count, black_count);
					white_count = 0;
					black_count = 0;
				}
				i--;
			}
		}

		// Rearrange results if black is playing and return results
		if (playing_colour == 'b')
		{
			int[] temp = new int[6];
			temp[0] = results[3];
			temp[1] = results[4];
			temp[2] = results[5];
			temp[3] = results[0];
			temp[4] = results[1];
			temp[5] = results[2];
			return temp;
		}
		else return results;
	}


	private void update_count(int[] results, int white_count, int black_count)
	// Used by count_contiguous method to update results when a contiguous string ends.
	// Results is the array to be updated, and white_count and black_count are the number
	// of contiguous blocks of each colour.
	// By default, white is considered the player colour at this stage (count_contiguous
	// will change results later if black is the player colour).
	{
		if (black_count == 0)
		{
			if (white_count == 2)
				results[0]++;
			else if (white_count == 3)
				results[1]++;
			else if (white_count >= 4)
				results[2]++;
		}
		else
		{
			if (black_count == 2)
				results[3]++;
			else if (black_count == 3)
				results[4]++;
			else if (black_count >= 4)
				results[5]++;
		}
	}


	private void set_up_children(Object[] children_boards, int[] children_scores, char[] moves,
	                             char[][] board, char computer_colour, boolean is_max)
	// Initialize children information for max_value or min_value procedures
	{
		// Set up the moves array
		moves[0] = '0';
		moves[1] = '1';
		moves[2] = '2';
		moves[3] = '3';
		moves[4] = '4';
		moves[5] = '5';
		moves[6] = '6';
		moves[7] = 'a';
		moves[8] = 'b';
		moves[9] = 'c';
		moves[10] = 'd';
		moves[11] = 'e';
		moves[12] = 'f';
		moves[13] = 'g';
		
		// Determine move_colour (the colour of the piece that will move in the next turn
		char move_colour;
		if (!is_max)
		{
			if (computer_colour == 'w') move_colour = 'b';
			else move_colour = 'w';
		}
		else move_colour = computer_colour;

		for (int i = 0; i < 14; i++)
		{
			children_boards[i] = copy_board(board);
			revise_state(moves[i], (char[][]) children_boards[i], move_colour);
			children_scores[i] = evaluate_board((char[][]) children_boards[i], computer_colour);
		}
	}


	private char check_legal_move(char[][] board, char given_move)
	// Ensures that move is not into a full row or column
	{
		boolean legal = false; // true if move is legal

		// Check if move is legal
		if (given_move == '0' && board[0][0] == ',') legal = true;
		else if (given_move == '1' && board[1][0] == ',') legal = true;
		else if (given_move == '2' && board[2][0] == ',') legal = true;
		else if (given_move == '3' && board[3][0] == ',') legal = true;
		else if (given_move == '4' && board[4][0] == ',') legal = true;
		else if (given_move == '5' && board[5][0] == ',') legal = true;
		else if (given_move == '6' && board[6][0] == ',') legal = true;
		else if (given_move == 'a' && board[6][0] == ',') legal = true;
		else if (given_move == 'b' && board[6][1] == ',') legal = true;
		else if (given_move == 'c' && board[6][2] == ',') legal = true;
		else if (given_move == 'd' && board[6][3] == ',') legal = true;
		else if (given_move == 'e' && board[6][4] == ',') legal = true;
		else if (given_move == 'f' && board[6][5] == ',') legal = true;
		else if (given_move == 'g' && board[6][6] == ',') legal = true;

		// Return the given move if it is legal
		if (legal) return given_move;
		
		// Cycle through other possible moves if it is not
		else
		{
			if (board[0][0] == ',') return '0';
			else if (board[1][0] == ',') return '1';
			else if (board[2][0] == ',') return '2';
			else if (board[3][0] == ',') return '3';
			else if (board[4][0] == ',') return '4';
			else if (board[5][0] == ',') return '5';
			else if (board[6][0] == ',') return '6';
			else if (board[6][0] == ',') return 'a';
			else if (board[6][1] == ',') return 'b';
			else if (board[6][2] == ',') return 'c';
			else if (board[6][3] == ',') return 'd';
			else if (board[6][4] == ',') return 'e';
			else if (board[6][5] == ',') return 'f';
			else return 'g';
		}
	}


	private char[][] copy_board(char[][] board)
	// Returns a copy of the given state space
	{
		char[][] temp = new char[7][7];
		for (int i = 0; i < 7; i++)
		{
			for (int j = 0; j < 7; j++)
				temp[i][j] = board[i][j];
		}
		return temp;
	}
}